"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.PgAttributesPlugin = void 0;
require("./PgTablesPlugin.js");
require("../interfaces.js");
require("graphile-config");
const pg_1 = require("@dataplan/pg");
const grafast_1 = require("grafast");
const graphile_build_1 = require("graphile-build");
const version_js_1 = require("../version.js");
function processAttribute(fields, build, context, attributeName, overrideName) {
    const { extend, inflection, input: { pgRegistry: registry }, } = build;
    const { scope: { pgCodec: rawPgCodec }, } = context;
    if (!rawPgCodec || !rawPgCodec.attributes) {
        return;
    }
    const pgCodec = rawPgCodec;
    const isInterface = context.type === "GraphQLInterfaceType";
    const attribute = pgCodec.attributes[attributeName];
    if (!build.behavior.pgCodecAttributeMatches([pgCodec, attributeName], "attribute:select")) {
        // Don't allow selecting this attribute.
        return;
    }
    const attributeFieldName = overrideName ??
        inflection.attribute({
            attributeName,
            codec: pgCodec,
        });
    const resolveResult = resolveOutputType(build, attribute.codec, attribute.notNull);
    if (!resolveResult) {
        console.warn(`Couldn't find a 'output' variant for PgCodec ${pgCodec.name}'s '${attributeName}' attribute (${attribute.codec.name}; array=${!!attribute.codec.arrayOfCodec}, domain=${!!attribute.codec
            .domainOfCodec}, enum=${!!attribute.codec.values})`);
        return;
    }
    const [baseCodec, type] = resolveResult;
    if (!type) {
        // Could not determine the type, skip this field
        console.warn(`Could not determine the type for attribute '${attributeName}' of ${pgCodec.name}`);
        return;
    }
    const fieldSpec = {
        description: attribute.description,
        type: type,
    };
    const executor = pgCodec.executor;
    const resource = baseCodec.attributes
        ? Object.values(registry.pgResources).find((potentialSource) => potentialSource.codec === baseCodec &&
            !potentialSource.parameters &&
            potentialSource.executor === executor)
        : null;
    if (baseCodec.attributes && !resource) {
        // We can't load codecs with attributes unless we know the executor.
        return;
    }
    if (!isInterface) {
        const makePlan = () => {
            // See if there's a resource to pull record types from (e.g. for relations/etc)
            if (!baseCodec.attributes) {
                // Simply get the value
                return (0, graphile_build_1.EXPORTABLE)((attributeName) => ($record) => {
                    return $record.get(attributeName);
                }, [attributeName]);
            }
            else {
                if (!resource) {
                    // This error only exists to satisfy TypeScript
                    throw new Error("This should be unreachable");
                }
                // TODO: this is pretty horrible in the export; we should fix that.
                if (!attribute.codec.arrayOfCodec) {
                    const notNull = attribute.notNull || attribute.codec.notNull;
                    // Single record from resource
                    /*
                     * TODO: if we refactor `PgSelectSingleStep` we can probably
                     * optimise this to do inline selection and still join against
                     * the base table using e.g. `(table.column).attribute =
                     * joined_thing.column`
                     */
                    return (0, graphile_build_1.EXPORTABLE)((attributeName, notNull, pgSelectSingleFromRecord, resource) => ($record) => {
                        const $plan = $record.get(attributeName);
                        const $select = pgSelectSingleFromRecord(resource, $plan);
                        if (notNull) {
                            $select.coalesceToEmptyObject();
                        }
                        $select.getClassStep().setTrusted();
                        return $select;
                    }, [attributeName, notNull, pg_1.pgSelectSingleFromRecord, resource]);
                }
                else if (attribute.codec.arrayOfCodec.arrayOfCodec?.arrayOfCodec) {
                    throw new Error("Triple nested arrays are currently unsupported... Feel free to send a PR that refactors this whole section!");
                }
                else if (attribute.codec.arrayOfCodec.arrayOfCodec) {
                    return (0, graphile_build_1.EXPORTABLE)((attributeName, each, pgSelectFromRecords, resource) => ($record) => {
                        const $val = $record.get(attributeName);
                        return each($val, ($list) => {
                            const $select = pgSelectFromRecords(resource, $list);
                            $select.setTrusted();
                            return $select;
                        });
                    }, [attributeName, grafast_1.each, pg_1.pgSelectFromRecords, resource]);
                }
                else {
                    // atrribute.codec.arrayOfCodec is set
                    // Many records from resource
                    return (0, graphile_build_1.EXPORTABLE)((attributeName, pgSelectFromRecords, resource) => ($record) => {
                        const $val = $record.get(attributeName);
                        const $select = pgSelectFromRecords(resource, $val);
                        $select.setTrusted();
                        return $select;
                    }, [attributeName, pg_1.pgSelectFromRecords, resource]);
                }
            }
        };
        fieldSpec.plan = makePlan();
    }
    fields = extend(fields, {
        [attributeFieldName]: context.fieldWithHooks({
            fieldName: attributeFieldName,
            pgFieldAttribute: attribute,
        }, fieldSpec),
    }, `Adding '${attributeName}' attribute field to PgCodec '${pgCodec.name}'`);
}
exports.PgAttributesPlugin = {
    name: "PgAttributesPlugin",
    description: "Adds PostgreSQL attributes (columns) to the relevant GraphQL object/input object types",
    version: version_js_1.version,
    after: ["PgTablesPlugin"],
    inflection: {
        add: {
            _attributeName(options, { attributeName, codec }) {
                const attribute = codec.attributes[attributeName];
                return this.coerceToGraphQLName(attribute.extensions?.tags?.name || attributeName);
            },
            _joinAttributeNames(options, codec, names) {
                return names
                    .map((attributeName) => {
                    return this.attribute({ attributeName, codec });
                })
                    .join("-and-");
            },
            attribute(options, details) {
                const attributeFieldName = this.camelCase(this._attributeName(details));
                // Avoid conflict with 'id' field used for Relay.
                return attributeFieldName === "id" && !details.codec.isAnonymous
                    ? "rowId"
                    : attributeFieldName;
            },
        },
    },
    schema: {
        entityBehavior: {
            pgCodecAttribute: "select base update insert",
        },
        hooks: {
            GraphQLInterfaceType_fields(fields, build, context) {
                const { scope: { pgCodec, pgPolymorphism }, } = context;
                if (!pgPolymorphism || !pgCodec?.attributes) {
                    return fields;
                }
                for (const attributeName in pgCodec.attributes) {
                    switch (pgPolymorphism.mode) {
                        case "single": {
                            if (!pgPolymorphism.commonAttributes.includes(attributeName)) {
                                continue;
                            }
                            break;
                        }
                        case "relational": {
                            break;
                        }
                        case "union": {
                            break;
                        }
                        default: {
                            const never = pgPolymorphism;
                            throw new Error(`Unhandled polymorphism mode ${never.mode}}`);
                        }
                    }
                    processAttribute(fields, build, context, attributeName);
                }
                return fields;
            },
            GraphQLObjectType_fields(fields, build, context) {
                const { scope: { pgCodec, isPgClassType, pgPolymorphism, pgPolymorphicSingleTableType, }, } = context;
                if (!isPgClassType || !pgCodec?.attributes) {
                    return fields;
                }
                for (const attributeName in pgCodec.attributes) {
                    let overrideName = undefined;
                    if (pgPolymorphism) {
                        switch (pgPolymorphism.mode) {
                            case "single": {
                                const match = pgPolymorphicSingleTableType?.attributes.find((c) => c.attribute === attributeName);
                                if (!pgPolymorphism.commonAttributes.includes(attributeName) &&
                                    !match) {
                                    continue;
                                }
                                if (match?.rename) {
                                    overrideName = match.rename;
                                }
                                break;
                            }
                            case "relational": {
                                break;
                            }
                            case "union": {
                                break;
                            }
                            default: {
                                const never = pgPolymorphism;
                                throw new Error(`Unhandled polymorphism mode ${never.mode}}`);
                            }
                        }
                    }
                    processAttribute(fields, build, context, attributeName, overrideName);
                }
                return fields;
            },
            GraphQLInputObjectType_fields(fields, build, context) {
                const { extend, inflection, input: { pgRegistry }, } = build;
                const { scope: { isPgRowType, isPgCompoundType, isPgPatch, isPgBaseInput, pgCodec: rawPgCodec, }, fieldWithHooks, } = context;
                if (!(isPgRowType || isPgCompoundType) ||
                    !rawPgCodec ||
                    !rawPgCodec.attributes ||
                    rawPgCodec.isAnonymous) {
                    return fields;
                }
                const pgCodec = rawPgCodec;
                return Object.entries(pgCodec.attributes).reduce((memo, [attributeName, attribute]) => build.recoverable(memo, () => {
                    const action = isPgBaseInput
                        ? "base"
                        : isPgPatch
                            ? "update"
                            : "insert";
                    const relations = Object.values(pgRegistry.pgRelations[pgCodec.name] ?? {}).filter((r) => !r.isReferencee && r.isUnique);
                    const isPartOfRelation = relations.some((r) => r.localAttributes.includes(attributeName));
                    const prefix = isPartOfRelation ? `relation:` : ``;
                    const fieldBehaviorScope = `${prefix}attribute:${action}`;
                    if (!build.behavior.pgCodecAttributeMatches([pgCodec, attributeName], fieldBehaviorScope)) {
                        return memo;
                    }
                    const fieldName = inflection.attribute({
                        attributeName,
                        codec: pgCodec,
                    });
                    if (memo[fieldName]) {
                        throw new Error(`Two attributes produce the same GraphQL field name '${fieldName}' on input PgCodec '${pgCodec.name}'; one of them is '${attributeName}'`);
                    }
                    const attributeType = build.getGraphQLTypeByPgCodec(attribute.codec, "input");
                    if (!attributeType) {
                        return memo;
                    }
                    return extend(memo, {
                        [fieldName]: fieldWithHooks({
                            fieldName,
                            fieldBehaviorScope,
                            pgCodec,
                            pgAttribute: attribute,
                        }, {
                            description: attribute.description,
                            type: build.nullableIf(isPgBaseInput ||
                                isPgPatch ||
                                (!attribute.notNull &&
                                    !attribute.extensions?.tags?.notNull) ||
                                attribute.hasDefault ||
                                Boolean(attribute.extensions?.tags?.hasDefault), attributeType),
                            autoApplyAfterParentInputPlan: true,
                            autoApplyAfterParentApplyPlan: true,
                            applyPlan: (0, graphile_build_1.EXPORTABLE)((attributeName) => function plan($insert, val) {
                                $insert.set(attributeName, val.get());
                            }, [attributeName]),
                        }),
                    }, `Adding input object field for ${pgCodec.name}.`);
                }), fields);
            },
        },
    },
};
function resolveOutputType(build, codec, notNull) {
    const { getGraphQLTypeByPgCodec, graphql: { GraphQLList, GraphQLNonNull, getNullableType, isOutputType }, } = build;
    if (codec.arrayOfCodec) {
        const resolvedResult = resolveOutputType(build, codec.arrayOfCodec);
        if (!resolvedResult) {
            return null;
        }
        const [innerCodec, innerType] = resolvedResult;
        const nullableType = new GraphQLList(innerType);
        const type = notNull || codec.notNull
            ? new GraphQLNonNull(nullableType)
            : nullableType;
        return [innerCodec, type];
    }
    else {
        const baseType = getGraphQLTypeByPgCodec(codec, "output");
        if (!baseType || !isOutputType(baseType)) {
            return null;
        }
        const type = notNull || codec.notNull
            ? new GraphQLNonNull(getNullableType(baseType))
            : baseType;
        return [codec, type];
    }
}
//# sourceMappingURL=PgAttributesPlugin.js.map